//
// Authors:
//   Patrick Cernko <cernko@ps.uni-sb.de>
//
// Copyright:
//   Patrick Cernko, 2004
//
// Last Change:
//   $Date: 2004/03/15 14:31:43 $ by $Author: errror $
//   $Revision: 1.1.2.2 $
//

#if defined(INTERFACE)
#pragma implementation "java/KaffeInterpreter.hh"
#endif

#include "KaffeInterpreterHelpers.hh"

//
// Interpreter StackFrames
//
#include "KaffeFrame.hh"

//
// Unlock Worker
//
#include "UnlockWorker.hh"

//
// Interpreter Functions
//
KaffeInterpreter *KaffeInterpreter::self;

void KaffeInterpreter::Init() {
  UnlockWorker::Init();
  self = new KaffeInterpreter();
}

Transform *
KaffeInterpreter::GetAbstractRepresentation(ConcreteRepresentation *) {
  return INVALID_POINTER; // to be done
}

void KaffeInterpreter::PushCall(Closure *closure) {
  JavaByteCode *concreteCode =
    JavaByteCode::FromWord(closure->GetConcreteCode());
  KaffeFrame::New(KaffeInterpreter::self,
                  -1,
                  0,
                  concreteCode,
                  closure->Sub(0)); // RuntimeConstantPool
}

u_int KaffeInterpreter::GetFrameSize(StackFrame *sFrame) {
  KaffeFrame *frame = STATIC_CAST(KaffeFrame *, sFrame);
  Assert(sFrame->GetWorker() == this);
  return frame->GetSize();
}

Interpreter::Result KaffeInterpreter::Handle(word) {
  StackFrame *sFrame = Scheduler::GetFrame();
  Assert(sFrame->GetWorker() == this);
  KaffeFrame *frame = STATIC_CAST(KaffeFrame *, sFrame);
  s_int pc             = frame->GetPC();
  Table *table         = frame->GetExceptionTable();
  u_int count          = table->GetCount();
  Object *object       = Object::FromWord(Scheduler::currentData);
  Assert(object != INVALID_POINTER);
  for (u_int i = 0; i < count; i++) {
    ExceptionTableEntry *entry =
      ExceptionTableEntry::FromWordDirect(table->Get(i));
    s_int startPC = entry->GetStartPC();
    s_int endPC   = entry->GetEndPC();
    // Exception handler is within range
    if ((startPC <= pc) && (pc < endPC)) {
      // Check exception type
      word wType = entry->GetCatchType();
      if (wType == null) {
	frame->SetPC(entry->GetHandlerPC());
	return Worker::CONTINUE;
      }
      Type *type = Type::FromWord(wType);
      Assert(type != INVALID_POINTER); //--** to be done: should request
      if (IsInstanceOf(object->ToWord(), type)) {
	frame->SetPC(entry->GetHandlerPC());
	return Worker::CONTINUE;
      }
    }
  }
  word wFrame = frame->Clone();
  Scheduler::PopFrame(frame->GetSize());
  Scheduler::currentBacktrace->Enqueue(wFrame);
  return Worker::RAISE;
}

u_int KaffeInterpreter::GetInArity(ConcreteCode *concreteCode) {
  Assert(concreteCode->GetInterpreter() == KaffeInterpreter::self);
  JavaByteCode *byteCode = STATIC_CAST(JavaByteCode *, concreteCode);
  u_int arity = byteCode->GetMethodInfo()->GetNumberOfArguments();
  return arity;
}

u_int KaffeInterpreter::GetOutArity(ConcreteCode *concreteCode) {
  Assert(concreteCode->GetInterpreter() == KaffeInterpreter::self);
  JavaByteCode *byteCode = STATIC_CAST(JavaByteCode *, concreteCode);
  u_int arity = byteCode->GetMethodInfo()->GetNumberOfReturnValues();
  return arity;
}

const char *KaffeInterpreter::Identify() {
  return "KaffeInterpreter";
}

void KaffeInterpreter::DumpFrame(StackFrame *sFrame) {
  KaffeFrame *frame = STATIC_CAST(KaffeFrame *, sFrame);
  Assert(sFrame->GetWorker() == this);
  MethodInfo *info     = frame->GetMethodInfo();
  std::fprintf(stderr, "%s#%s%s\n", info->GetClassName()->ExportC(),
	       info->GetName()->ExportC(), info->GetDescriptor()->ExportC());
}

void KaffeInterpreter::FillStackTraceElement(word wFrame, 
						Object *stackTraceElement) {
  // to be done: Hack Alert
  u_int size = Store::DirectWordToBlock(wFrame)->GetSize();
  StackFrame *sFrame = Scheduler::PushFrame(size);
  StackFrame::New(sFrame, size, wFrame);
  KaffeFrame *frame = STATIC_CAST(KaffeFrame *, sFrame);
  Assert(sFrame->GetWorker() == KaffeInterpreter::self);
  JavaString *className = frame->GetMethodInfo()->GetClassName();
  JavaString *name = frame->GetMethodInfo()->GetName();
  stackTraceElement->InitInstanceField
    (StackTraceElement::DECLARING_CLASS_INDEX, className->ToWord());
  stackTraceElement->InitInstanceField
    (StackTraceElement::METHOD_NAME_INDEX, name->ToWord());
  stackTraceElement->InitInstanceField
    (StackTraceElement::FILE_NAME_INDEX, null);
  stackTraceElement->InitInstanceField
    (StackTraceElement::LINE_NUMBER_INDEX, JavaInt::ToWord(-1));
}
