//
// Authors:
//   Patrick Cernko <cernko@ps.uni-sb.de>
//
// Copyright:
//   Patrick Cernko, 2004
//
// Last Change:
//   $Date: 2004/09/06 13:00:57 $ by $Author: errror $
//   $Revision: 1.1.2.2 $
//

#include <cstdio> // fprintf()
#include <cmath>  // floor()

#include "java/Opcodes.hh"
#include "java/ThrowWorker.hh"
#include "java/KaffeInterpreter.hh"
#include "java/KaffeFrame.hh"
#include "java/UnlockWorker.hh"
#include "java/ConstantTag.hh"

//
// Helper Stuff
//
#include "KaffeInterpreterHelpers.hh"

#undef REQUEST
#define REQUEST(w) {	      \
  dprintf("REQUEST: pc=%d\n", \
          pc);                \
  frame->SetPC(pc);           \
  frame->SetTop(frameStackTop); \
  Scheduler::currentData = w; \
  Scheduler::nArgs = 0;	      \
  return Worker::REQUEST;     \
}

#define CHECK_PREEMPT() {			\
  if (StatusWord::GetStatus() != 0) {		\
    return Worker::PREEMPT;			\
  } else {					\
    return Worker::CONTINUE;			\
  }                                             \
}

#define RAISE_EXCEPTION(exn) { \
  word wFrame = frame->Clone(); \
  Scheduler::PopFrame(frame->GetSize()); \
  Scheduler::currentData = exn; \
  Scheduler::currentBacktrace = Backtrace::New(wFrame); \
  return Worker::RAISE; \
}

#define RAISE_VM_EXCEPTION(exn, mesg) { \
  ThrowWorker::PushFrame(ThrowWorker::exn, JavaString::New(mesg)); \
  Scheduler::nArgs = 0; \
  return Worker::CONTINUE; \
}

//
// Interpreter Types
//
// to be done: check compliance of java spec with c spec
class JavaFloat {
public:
  static word ToWord(double value) {
    return Float::New(value)->ToWord();
  }
  static Float *FromWord(word value) {
    return Float::FromWordDirect(value);
  }
};

// to be done: check compliance of java spec with c spec
class JavaDouble {
public:
  static word ToWord(double value) {
    return Double::New(value)->ToWord();
  }
  static Double *FromWord(word value) {
    return Double::FromWordDirect(value);
  }
};



#define INTERPRETER 1
#define HAVE_BOOL 1

extern "C" {
#include "config.h"
#include "config-std.h"
#include "config-math.h"

#include "debug.h"
#include "kaffe/jtypes.h"

/* gtypes.h defines a EXIT, but don't use it as the one seam defined! */
#undef EXIT
#include "gtypes.h"

// #include "slots.h"
// #include "icode.h"
// #include "checks.h"
#include "soft.h"
}

#include "DefaultKaffeAccessors.hh"
#include "KaffeAccessors.hh"

extern "C" {
#include "code-analyse.h"
}

Worker::Result KaffeInterpreter::Run(StackFrame *sFrame) {
  KaffeFrame *frame = STATIC_CAST(KaffeFrame *, sFrame);
  Assert(sFrame->GetWorker() == this);
  s_int pc = frame->GetPC();
  u_int8 *code = reinterpret_cast<u_int8 *>(frame->GetCode()->GetBase());
  RuntimeConstantPool *pool = frame->GetRuntimeConstantPool();
  word* frameStackBase = frame->GetBase();
  u_int frameStackTop = frame->GetTop();

#ifdef DEBUG_LEVEL3
  fprintf(stderr,
          "KaffeInterpreter::Run()"
          " Frame: { method=\"%s::%s()\" (%s), id=%d, pc=%d, code=%p, pool=%p, cont_pc=%d }\n",
          frame->GetMethodInfo()->GetClassName()->ExportC(),
          frame->GetMethodInfo()->GetName()->ExportC(),
          frame->GetMethodInfo()->GetDescriptor()->ExportC(),
#ifdef DEBUG_KAFFEFRAMEID
          frame->GetUniqueID(), 
#else
          -1,
#endif
          pc, code, pool, frame->GetContPC());
#endif

#ifdef DEBUG_LEVEL5
  frame->Dump();
#endif

  // to be done: more efficient solution
  if (pc == -1) {
    // Copy arguments to local variables
    // to be done: support large arguments
    for (u_int i = Scheduler::nArgs; i--;) {
      *(local(i)) = Scheduler::currentArgs[i];
    }
    if (frame->GetExceptionTable()->GetCount() != 0) {
//       fprintf(stderr, 
//               "Scheduler::PushHandler() in KaffeInterpreter::Run(): currentThread=%p, data=%p\n", 
//               Scheduler::GetCurrentThread(), 
//               (void*) Store::IntToWord(0));
      Scheduler::PushHandler(Store::IntToWord(0));
    }
    pc = frame->GetContPC();
    Scheduler::nArgs = 0;
  }

  else if (pc == -2) {
    // Push Return Value to stack
    for (u_int i = 0; i < Scheduler::nArgs; i++) {
      push(1);
      *(stack(0)) = Scheduler::currentArgs[i];
    }
    if (frame->GetExceptionTable()->GetCount() != 0) {
//       fprintf(stderr, 
//               "Scheduler::PopHandler() in KaffeInterpreter::Run(): currentThread=%p\n", 
//               Scheduler::GetCurrentThread());
      // XXX: Why poping Handler HERE???
      //      Scheduler::PopHandler();
    }
    pc = frame->GetContPC();
    Scheduler::nArgs = 0;
  }

  { // void* virtualMachine(methods*volatile meth, slots* volatile arg, slots* volatile retval, threadData* volatile thread_data)
    uintp volatile npc = pc;
    word retval = INVALID_POINTER;

    { //runVirtualMachine(meth, lcl, sp, npc, retval, &mjbuf, thread_data);
      /* Misc machine variables */
      jvalue tmpl;
      word tmp[1];
      word tmp2[1];
      word mtable[1];

      /* Variables used directly in the machine */
      int32 idx;
      jint low;
      jint high;

      /// XXX: do we need this?
      /* Call, field and creation information */
      callInfo cinfo;
      fieldInfo finfo;
      Hjava_lang_Class* crinfo;
      Hjava_lang_Class class_struct;

      StaticFieldRef* staticFieldRef = NULL;
      InstanceFieldRef* instanceFieldRef = NULL;
      Class* fieldClass = NULL;
      Lock* fieldClassLock = NULL;

      StaticMethodRef* staticMethodRef = NULL;
      VirtualMethodRef* virtualMethodRef = NULL;
      Class* methodClass = NULL;
      Lock* methodClassLock = NULL;

      /* Finally we get to actually execute the machine */
      while (true) {
        register uintp pc = npc;
        // XXX: to be done
//         assert(npc < meth->c.bcode.codelen);
        npc = pc + insnLen[code[pc]];

        switch (code[pc]) {
        default:
          dprintf("Unknown bytecode %d\n", code[pc]);
          // XXX: to be done/do we need this?
//           throwException(NEW_LANG_EXCEPTION(VerifyError));
          break;
#include "kaffe.def"
        }
        
        // XXX: check preempt!!!
        if (StatusWord::GetStatus() != 0) {
          frame->SetPC(npc);
          frame->SetTop(frameStackTop);
          return Worker::PREEMPT;
        }
      }
    } // END runVirtualMachine(meth, lcl, sp, npc, retval, &mjbuf, thread_data);

    return Worker::CONTINUE;
  }
}
